'''
This is the "InputMotor" class. It inherits from "Component" class.
The input motor class defines the properties of an electric motor based on:
I) Given properties:
--> 1) "name": a string of characters acting as a label
--> 2) "material": a "Material" object of the material class defining the material properties of the component
--> 3) "axis": a 3-element vector representing the axis along which the motor is rotating with respect to a defined reference frame
--> 4) "loc": a 3-element vector representing the location of the motor with respect to a defined reference frame
--> 5) "power": scalar value representing the power of the motor expressed in [W]
--> 6) "n": scalar value representing the speed expressed in [rpm]
II) Calculated properties
--> 1) "omega": a 3-vector element representing the rotational velocity of the vector expressed in [rad/s]
--> 2) "F_tot": the total force acting on the motor expressed in [N] (initialized to 0)
--> 3) "T_tot": the total torque acting on the motor expressed in [Nm] (based on power and speed)
'''
import numpy as np
from .component import Component
from .force import Force
from .torque import Torque
from math import pi

class InputMotor(Component):
    
    # Constructor
    def __init__(self, name, power, n, axis, loc):
        # Given properties
        super().__init__(name=name, material=None, axis=axis, loc=loc, F_tot=None, T_tot=None, omega=None)
        self.power = power
        self.n = n
        # Calculated properties
        self.omega = self.n * pi / 30 * self.axis
        self.T_tot = Torque(np.array([0 if o == 0 else self.power/o for o in self.omega]), self.loc)
        self.F_tot = Force(np.array([0, 0, 0]), self.loc)
        



import numpy as np
from math import pi
class ShaftProfile:

    # Constructor
    def __init__(self, radii, locs):
        self.radii = np.concatenate(([0], radii, [0]))
        self.locs = np.concatenate(([locs[0]], locs, [locs[-1]]))
    
    # Add Fillet
    def AddFillet(self, radius, quadrant, zOff, dOff):
        for q in range(len(quadrant)):
            if quadrant[q] == 1:
                theta = np.arange(pi, 3*pi/2, 0.1)
            elif quadrant[q] == 2:
                theta = np.arange(3*pi/2, 2*pi, 0.1)
            else:
                raise ValueError("Wrong Input")
        z = radius * np.cos(theta) + zOff
        r = radius * np.sin(theta) + dOff
        before = self.locs < np.min(z)
        after = self.locs > np.max(z)
        self.radii = np.concatenate((self.radii[np.where(before)], r, self.radii[np.where(after)]))
        self.locs = np.concatenate((self.locs[np.where(before)], z, self.locs[np.where(after)]))
    
    # Refine Profile
    def refineProfile(self, delta):
        pLen = int((self.locs[-1] - self.locs[0]) / delta + 1)
        refinedProfile = ShaftProfile(np.zeros(pLen), np.arange(self.locs[0], self.locs[-1] + delta / 2, delta))
        for z in range(2, len(self.locs) - 3):
            condition = np.where(np.logical_and(refinedProfile.locs >= self.locs[z], refinedProfile.locs <= self.locs[z + 1]))
            if self.locs[z] != self.locs[z + 1]:
                refinedProfile.radii[condition] = np.interp(refinedProfile.locs[condition], np.array([self.locs[z], self.locs[z+1]]), np.array([self.radii[z], self.radii[z+1]]))
        return refinedProfile
    
    # Calculate Cross-Sectional Properties
    def CalculateSectionProperties(self):
        self.Area = pi * self.radii ** 2
        self.Wb = pi / 4 * self.radii ** 3
        self.Wt = pi / 2 * self.radii ** 3
    
    # Plot Profile
    def plotProfile(self, ax):
        ax1 = ax.twinx()
        ax1.plot(self.locs, self.radii, 'r', linewidth=1.5)
        ax1.plot(self.locs, -self.radii, 'r', linewidth=1.5)
        window = (self.locs[-1] - self.locs[0] + 20) / 2
        ax1.set_xlim(-0.1 * window, 2 * window)
        ax1.set_ylim(-window, window)
        ax1.set_ylabel("Profile [mm]")


'''
This is the "Torque" class.
It defines a torque vector based on two simple properties:
1) "torque": a 3-element torque vector representing the torque expressed in [Nm]
2) "loc": a 3-element vector representing the point of application of the torque expressed in [mm]
'''
class Torque:

    # Constructor
    def __init__(self, torque, loc):
        self.torque = torque
        self.loc = loc
    
    # Overload Addition
    def __add__(obj1, obj2):
        return obj1.torque + obj2.torque
    
    # Overload Subtraction
    def __sub__(obj1, obj2):
        return obj1.torque - obj2.torque
    
    # Overload Negative
    def __neg__(obj):
        return -obj.torque
    
        # Overload Equal
    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return all(self.torque == other.torque) and all(self.loc == other.loc)
        return False
    
    # Overload Call
    def __call__(self):
        print(f"Torque: {self.torque}\nLoc: {self.loc}")



'''
This is the NotchSensitivity class.
It contains the data and the necessary calculations via interpolation to get the final factor.
'''
import numpy as np
import matplotlib.pyplot as plt
from .makima2dInterpolator import Makima2DInterpolator
class NotchSensitivity:
    # Notch radius [mm]
    notchRadius = np.arange(0.25, 4.1, 0.25)
    # Ultimate Strength [MPa]
    sigma_u = np.array([.4, .7, 1, 1.4])*1e3
    # Notch Sensitivity
    qRef = np.array([[0.48, 0.57, 0.62, 0.65, 0.68, 0.7, 0.72, 0.73, 0.74, 0.755, 0.76, 0.77, 0.775, 0.78, 0.785, 0.79],
                    [0.595, 0.7, 0.75, 0.78, 0.79, 0.8, 0.81, 0.82, 0.83, 0.84, 0.845, 0.85, 0.855, 0.86, 0.865, 0.87],
                    [0.74, 0.81, 0.85, 0.87, 0.885, 0.892, 0.895, 0.897, 0.899, 0.903, 0.905, 0.908, 0.912, 0.915, 0.92, 0.922],
                    [0.85, 0.9, 0.92, 0.94, 0.946, 0.952, 0.96, 0.964, 0.97, 0.973, 0.975, 0.9755, 0.976, 0.9763, 0.9765, 0.9767]])
    # Makima Interpolation Function
    interp_func = Makima2DInterpolator(notchRadius, sigma_u, qRef)
    
    # Constructor
    def __init__(self, notchRadius, sigma_u):
        self.qReq = self.__class__.interp_func(notchRadius, sigma_u)
    
    # Plot Reference Notch Sensitivity
    def plotqRef(self):
        plt.figure
        for i in range(len(self.sigma_u)):
            plt.plot(self.notchRadius, self.qRef[i], linewidth=1.5, label="Su = " + str(self.sigma_u[i]) + "[MPa]")
        plt.xlabel("Notch Radius [mm]")
        plt.ylabel("q [-]")
        plt.title("Notch Sensitivity")
        plt.legend()
        plt.grid()
        plt.show()


'''
This is the Makima2DInterpolator class.
It allows to define an interpolator based on the Modified Akima method.
A linear extrapolation method is used across the bounds.
'''
import numpy as np
from scipy.interpolate import PchipInterpolator, Akima1DInterpolator

class Makima2DInterpolator:

    # Constructor
    def __init__(self, x, y, z):
        self.x = np.asarray(x)
        self.y = np.asarray(y)
        self.z = np.asarray(z)

        if self.z.shape != (len(y), len(x)):
            raise ValueError(f"Z shape must be ({len(y), len(x)}), instead got {self.z.shape}")

    
    def _akima_with_extrapolation(self, x, y_vals):
        # Return an interpolator with linear extrapolaion beyond bounds
        akima = Akima1DInterpolator(x, y_vals, extrapolate=True)
        def extrap(xq):
           xq = np.asarray(xq)
           result = akima(xq)
           nfit = 16
        #    coeffs = np.polyfit(x, y_vals, deg=3)
           # Left extrapolation
           left = xq < x[0]
           if np.any(left):
               coeffs = np.polyfit(x[:nfit], y_vals[:nfit], deg=2)
               result[left] = np.polyval(coeffs, xq[left])
                # slope = (y_vals[1] - y_vals[0]) / (x[1] - x[0])
                # result[left] = y_vals[0] + slope * (xq[left] - x[0])
            
            # Right extrapolation
           right = xq > x[-1]
           if np.any(right):
               coeffs = np.polyfit(x[-nfit:], y_vals[-nfit:], deg=2)
               result[right] = np.polyval(coeffs, x[right])
            #    slope = (y_vals[-1] - y_vals[-2]) / (x[-1] - x[-2])
            #    result[right] = y_vals[-1] + slope * (xq[right] - x[-1])
            
           return result
        return extrap
    
    # Call
    def __call__(self, xq, yq):

        # Interpolate along x (axis=1) for each value of y
        intermediate = []
        for row in self.z:
            interp = self._akima_with_extrapolation(self.x, row)
            intermediate.append(interp(xq))
        intermediate = np.array(intermediate)

        # Interpolate along y (axis=0)
        interp_final = self._akima_with_extrapolation(self.y, intermediate)
        return float(interp_final(yq))


import numpy as np
import matplotlib.pyplot as plt
from .component import Component
from .torque import Torque
from .force import Force
from math import pi
class Shaft(Component):

    # Constructor
    def __init__(self, name, inputs, outputs, axis, material, sup1, sup2, ids):
        # Given parameters
        super().__init__(name=name, material=material, axis=axis, loc=None, F_tot=None, T_tot=None, omega=inputs.omega)
        self.inputs = inputs
        self.outputs = outputs
        self.supports = np.array([sup1, sup2])
        self.ids = ids
        # Calculated parameters
        self.outputs.omega = self.omega
        self.sections = np.array([])
        sup1.omega = self.omega
        sup1.n = sup1.omega * 30 / pi
        sup2.omega = self.omega
        sup2.n = sup2.omega * 30 / pi
        self.EFs = np.array([])
        self.ETs = np.array([])
    
    # Calculate shaft torque
    def getShaftTorque(self):
        return Torque(-self.inputs.T_tot, self.outputs.loc * np.abs(self.axis))
    
    # Get shaft rotational speed
    def getOmegaShaft(self):
        return self.omega
    
    # Update external forces
    def updateEFs(self, EFs):
        for ef in EFs:
            if ef not in self.EFs:
                self.EFs = np.append(self.EFs, ef)
    
    # Update external torques
    def updateETs(self, ETs):
        for et in ETs:
            if et not in self.ETs:
                self.ETs = np.append(self.ETs, et)
    
    # Set shaft profile
    def setProfile(self, profile):
        self.profile = profile
    
    # Add sections
    def addSections(self, sections):
        for section in sections:
            section.material = self.material
        self.sections = np.append(self.sections, sections)
    
    # Insert fatigue limit corrector factors
    def insertFLCF(self, sections):
        for sec1, sec2 in zip(self.sections, sections):
            sec1.AddFLCF()
            sec2.AddFLCF()
    
    # Calculate reaction forces
    def calculateReactionForces(self):
        # Gears axial load
        K_a = np.zeros(3)
        for EF in self.EFs:
            K_a = K_a + (EF.force * self.axis)
        # Find the bearing around which to apply moment
        index = 0
        for i in range(len(self.supports)):
            if self.supports[i].type == "Pin":
                index = i
                self.supports[index].F_tot = Force(np.zeros(3), self.supports[index].loc * np.abs(self.axis))
        # Calculate reaction on other bearing
        for i in range(len(self.supports)):
            if i != index:
                self.supports[i].F_tot = Force(np.zeros(3), self.supports[i].loc)
                supDist = np.sum(self.supports[i].loc - self.supports[index].loc)
                for j in range(len(self.EFs)):
                    momEF = np.cross(self.EFs[j].force, self.EFs[j].loc - self.supports[index].loc) * 1e-3
                    self.supports[i].F_tot.force = self.supports[i].F_tot.force + np.cross(momEF, np.abs(self.axis) / supDist) * 1e3
                self.EFs = np.append(self.EFs, self.supports[i].F_tot)
        # Calculate reaction around bearing with sum of external forces
        for i in range(len(self.EFs)):
            self.supports[index].F_tot.force = self.supports[index].F_tot.force - self.EFs[i].force
        self.EFs = np.append(self.EFs, self.supports[index].F_tot)
        # Update axial load based on configuration
        if self.supports[0].bearingType == "Tapered" and self.supports[1].bearingType == "Tapered":
            if self.supports[0].shoulder == -1:
                indA = 1
                indB = 0
                sgn = 1
            else:
                indA = 0
                indB = 1
                sgn = -1
            A_FrV = self.supports[indA].F_tot.force - self.supports[indA].F_tot.force * self.axis
            A_Fr = np.sqrt(np.sum(A_FrV * A_FrV))
            A_Y = self.supports[indA].Y
            B_FrV = self.supports[indB].F_tot.force - self.supports[indB].F_tot.force * self.axis
            B_Fr = np.sqrt(np.sum(B_FrV * B_FrV))
            B_Y = self.supports[indB].Y
            # Case 1
            if np.sum(K_a) > 0:
                # Factor of comparison
                fac = B_Fr / B_Y - A_Fr / A_Y
                # Case 1a
                if fac <= 0 and np.abs(np.sum(K_a)) >= 0:
                    print("Case 1a")
                    A_Fa = sgn * 0.5 * A_Fr / A_Y * self.axis
                    B_Fa = -(A_Fa + K_a)
                # Case 1b
                elif fac > 0 and np.abs(np.sum(K_a)) >= 0.5 * fac:
                    print("Case 1b")
                    A_Fa = sgn * 0.5 * A_Fr / A_Y * self.axis
                    B_Fa = -(A_Fa + K_a)
                # Case 1c
                elif fac > 0 and np.abs(np.sum(K_a)) < 0.5 * fac:
                    print("Case 1c")
                    B_Fa = -sgn * 0.5 * B_Fr / B_Y * self.axis
                    A_Fa = -(B_Fa + K_a)
            # Case 2
            else:
                # Factor of comparison
                fac = A_Fr / A_Y - B_Fr / B_Y
                # Case 2a
                if fac <= 0 and np.abs(np.sum(K_a)) >= 0:
                    print("Case 2a")
                    B_Fa = -sgn * 0.5 * B_Fr / B_Y * self.axis
                    A_Fa = -(B_Fa + K_a)
                # Case 2b
                elif fac > 0 and np.abs(np.sum(K_a)) >= 0.5 * fac:
                    print("Case 2b")
                    B_Fa = -sgn * 0.5 * B_Fr / B_Y * self.axis
                    A_Fa = -(B_Fa + K_a)
                # Case 2c
                elif fac > 0 and np.abs(np.sum(K_a)) < 0.5 * fac:
                    print("Case 2c")
                    A_Fa = sgn * 0.5 * A_Fr / A_Y * self.axis
                    B_Fa = -sgn*(A_Fa + K_a)
            self.supports[indA].F_tot.force[2] = np.sum(A_Fa)
            self.supports[indB].F_tot.force[2] = np.sum(B_Fa)
            if self.supports[0].arr == "B2B":
                self.EFs[-2].force[2] = np.sum(B_Fa)
                self.EFs[-1].force[2] = np.sum(A_Fa)
            elif self.supports[0].arr == "F2F":
                self.EFs[-2].force[2] = np.sum(A_Fa)
                self.EFs[-1].force[2] = np.sum(B_Fa)
        # Check for equilibrium
        eq = np.zeros(3)
        for EF in self.EFs:
            eq = eq + EF.force
        if all(eq <= 1e-3 * np.ones(3)):
            print(f"{self.name} is in equilibrium")
        else:
            print(f"{self.name} is NOT in equilibrium")
    
    # Calculate internal loads
    def calculateInternalLoads(self, RF):
        l = len(self.profile.locs)
        self.N = np.zeros(l)
        self.Mx = np.zeros(l)
        self.My = np.zeros(l)
        self.Mt = np.zeros(l)
        for EF in self.EFs:
            for i, z in enumerate(self.profile.locs):
                if np.dot(EF.loc, np.abs(self.axis)) <= z:
                    self.N[i] = self.N[i] - np.sum(EF.force * np.abs(self.axis))
                    mxz = np.sum(np.cross(EF.force * RF[2], EF.loc * RF[1]))
                    mxy = np.sum(np.cross(EF.force * RF[1], (z - EF.loc ) * RF[2]))
                    self.Mx[i] = self.Mx[i] + (mxz - mxy) * 1e-3
                    myz = np.sum(np.cross(EF.force * RF[2], EF.loc * RF[0]))
                    myx = np.sum(np.cross(EF.force * RF[0], (EF.loc - z) * RF[2]))
                    self.My[i] = self.My[i] + (myz + myx) * 1e-3
        for ET in self.ETs:
                for i, z in enumerate(self.profile.locs):
                    if np.dot(ET.loc, np.abs(self.axis)) <= z:
                        self.Mt[i] = self.Mt[i] + np.sum(ET.torque)
        self.N[np.where(np.abs(self.N) < 1e-3)] = 0
        self.Mx[np.where(np.abs(self.Mx) < 1e-3)] = 0
        self.My[np.where(np.abs(self.My) < 1e-3)] = 0
        self.Mt[np.where(np.abs(self.Mt) < 1e-3)] = 0
        self.Mf = np.sqrt(self.Mx ** 2 + self.My ** 2)
    
    # Calculate stresses
    def calculateStresses(self):
        sLen = len(self.profile.locs)
        self.sigma_N = np.zeros(sLen)
        self.sigma_Mb = np.zeros(sLen)
        self.tau_Mt = np.zeros(sLen)
        self.sigma_N[np.where(self.profile.Area != 0)] = self.N[np.where(self.profile.Area != 0)] / self.profile.Area[np.where(self.profile.Area != 0)]
        self.sigma_Mb[np.where(self.profile.Wb != 0)] = 1e3 * self.Mf[np.where(self.profile.Wb != 0)] / self.profile.Wb[np.where(self.profile.Wb != 0)]
        self.tau_Mt[np.where(self.profile.Wt != 0)] = 1e3 * self.Mt[np.where(self.profile.Wt != 0)] / self.profile.Wt[np.where(self.profile.Wt != 0)]
    
    # Calculate equivalent and ideal stresses
    def calculateEquivalentAndIdealStress(self):
        self.sigma_tot = self.sigma_N + self.sigma_Mb
        self.sigma_id = np.sqrt(self.sigma_tot ** 2 + 3 * self.tau_Mt ** 2)
    
    # Plot internal loads
    def plotInternalLoads(self):
        # Normal load
        self.plotLoad(self.N, "N [N]", "Normal Load - N(z)")
        # Bending moment around x-axis
        self.plotLoad(self.Mx, r"$M_{x}$ [Nm]", r"Bending Moment $M_{x}(z)$")
        # Bending moment around y-axis
        self.plotLoad(self.My, r"$M_{y}$ [Nm]", r"Bending Moment $M_{y}(z)$")
        # Resulting bending moment
        self.plotLoad(self.Mf, r"$M_{B}$ [Nm]", r"Bending Moment $M_{B}(z)$")
        # Torsional moment
        self.plotLoad(self.Mt, r"$M_{t}$ [Nm]", r"Torsional Moment $M_{t}(z)$")
    
    # Plot load with shaft profile
    def plotLoad(self, load, ylabel, title):
        fig, ax = plt.subplots()
        ax.plot(self.profile.locs, load, 'b', linewidth = 1.5)
        ax.set_xlabel("z [mm]")
        ax.set_ylabel(ylabel)
        plt.title(title)
        if np.max(np.abs(load)) != 0:
            ax.set_ylim(-1.1 * np.max(np.abs(load)), 1.1 * np.max(np.abs(load)))
        else:
            ax.set_ylim(-1, 1)
        plt.grid()
        self.profile.plotProfile(ax)
        for section in self.sections:
            xs = np.ones(2) * np.sum(section.loc)
            ys = np.array([-1, 1])
            if np.max(np.abs(load)) != 0:
                ys = ys * 1.1 * np.max(np.abs(load))
            ax.plot(xs, ys, 'g--', linewidth=1.5)
            ax.text(xs[0] - 10, 0.9 * ys[0], section.name)
            ax.text(xs[0] - 10, 0.9 * ys[1], section.name)
    
    # Plot stresses
    def plotStresses(self):
        # Normal stress
        self.plotLoad(self.sigma_N, r"$\sigma^{N}$ [MPa]", r"Normal Stress - $\sigma^{N}(z)$ [MPa]")
        # Bending stress
        self.plotLoad(self.sigma_Mb, r"$\sigma^{M_{B}}$ [MPa]", r"Bending Stress - $\sigma^{M_{B}}(z)$ [MPa]")
        # Torsional stress
        self.plotLoad(self.tau_Mt, r"$\tau^{M_{t}}$ [MPa]", r"Torsional Stress - $\tau^{M_{t}}(z)$ [MPa]")
        # Total stress
        self.plotLoad(self.sigma_tot, r"$\sigma^{tot}$ [MPa]", r"Resulting Normal Stress - $\sigma^{tot}(z)$ [MPa]")
        # Equivalent stress
        self.plotLoad(self.sigma_id, r"$\sigma_{id}$ [MPa]", r"Equivalent Stress - $\sigma_{id}(z)$ [MPa]")
    
    # Calculate sections static safety factor
    def calculateStaticSafetyFactor(self, sections):
        for i in range(len(self.sections)):
            zV = np.sum(self.sections[i].loc)
            for j in range(len(self.profile.locs)):
                if zV >= self.profile.locs[j] and zV < self.profile.locs[j + 1]:
                    self.sections[i].staticSF = self.material.sigma_y / self.sigma_id[j]
            sections[i].staticSF = self.sections[i].staticSF

    # Calculate mean and alternating stresses
    def calculateMeanAlternatingStress(self, sections):
        for i in range(len(self.sections)):
            zV = np.sum(self.sections[i].loc)
            for j in range(len(self.profile.locs)):
                if zV >= self.profile.locs[j] and zV < self.profile.locs[j + 1]:
                    self.sections[i].sigma_m_N = self.sigma_N[j]
                    sections[i].sigma_m_N = self.sigma_N[j]
                    self.sections[i].sigma_a_N = 0
                    sections[i].sigma_a_N = 0
                    self.sections[i].sigma_m_Mb = 0
                    sections[i].sigma_m_Mb = 0
                    self.sections[i].sigma_a_Mb = self.sigma_Mb[j]
                    sections[i].sigma_a_Mb = self.sigma_Mb[j]
                    self.sections[i].tau_m_Mt = self.tau_Mt[j]
                    sections[i].tau_m_Mt = self.tau_Mt[j]
                    self.sections[i].tau_a_Mt = 0
                    sections[i].tau_a_Mt = 0
            sections[i].staticSF = self.sections[i].staticSF
    
    # Calculate equivalent mean and alternating stress
    def calculateEquivalentStresses(self, sections):
        for section in sections:
            section.CalculateEquivalentStress()


'''
This is the "HelicalGear" class. It inherits from Component class.
This class is responsible for defining the properties of a HelicalGear object, such as:
I) Given Properties
--> 1) "name": a string of characters acting as a label
--> 2) "axis": a 3-element vector representing the axis along which the motor is rotating with respect to a defined reference frame
--> 3) "loc": a 3-element vector representing the location of the motor with respect to a defined reference frame
--> 4) "m_n": normal module of the gear expressed in [mm]
--> 5) "z": number of teeth of the gear
--> 6) "psi": helix angle expressed in [rad]
--> 7) "phi_n": normal pressure angle expressed in [rad]
--> 8) "Q_v": transmission accuracy
--> 9) "FW": tooth face width expressed in [mm]
--> 10) "material": a Material object representing the material properties of the gear
II) Calculated parameters
--> 1) "p_n": npmral circular pitch expressed in [mm]
--> 2) "p_t": transverse circular pitch expressed in [mm]
--> 3) "p_x": axial pitch expressed in [mm]
--> 4) "m_t": transverse module expressed in [mm]
--> 5) "d": pitch diameter expressed in [mm]
--> 6) "phi_t": transverse pressure angle expressed in [rad]
--> 7) "z_p": virtual number of teeth
--> 8) "phi_b": base helix angle expressed in [rad]
--> 9) "h_a": addendum height expressed in [mm]
--> 10) "h_f": dedendum height expressed in [mm]
--> 11) "h": tooth height expressed in [mm]
--> 12) "d_a": addendum diameter expressed in [mm]
--> 13) "d_f": dedendum diameter expressed in [mm]
III) Calculated loads
--> 1) "F_t": a 3-element vector representing the tangential force expressed in [N]
--> 2) "F_r": a 3-element vector representing the radial force expressed in [N]
--> 3) "F_a": a 3-element vector representing the axial force expressed in [N]
IV) Gear tooth verification parameters
--> 1) "sigma_max_fatigue": maximum gear tooth bending stress for fatigue expressed in [MPa]
--> 2) "power_source": a string of characters deining the power source type
--> 3) "K_0": overload factor
--> 4) "t_R": difference between dedendum radius and internal gear radius expressed in [mm]
--> 5) "m_B": ratio between t_R and tooth height
--> 6) "K_B": rim-thickness factor
--> 7) "K_v": dynamic factor
--> 8) "C_mc": coefficient 1 for load distribution factor
--> 9) "C_pf": coefficient 2 for load distribution factor
--> 10) "C_pm": coefficient 3 for load distribution factor
--> 11) "C_ma": coefficient 4 for load distribution factor
--> 12) "C_e":coefficient 5 for load distribution factor
--> 13) "Y": coefficient 6 for load distribution factor
--> 14) "K_H": load distribution factor
--> 15) "K_S": size factor
--> 16) "J_p": geometry factor
--> 17) "J_pp": modifying factor
--> 18) "Y_J": bending strength geometry factor
--> 19) "sigma_FP": bending fatigue strength expressed in [MPa]
--> 20) "Y_N": stress cycle life factor (bending)
--> 21) "Y_theta": temperature factor
--> 22) "Y_Z": reliability factor
--> 23) "bendingSF": bending safety factor
--> 24) "sigma_max_pitting": maximum gear contact (pitting resistance) stess expressed in [MPa]
--> 25) "Z_E": elastic coefficient expressed in [MPa^(1/2)]
--> 26) "Z_I": surface strength geometry
--> 27) "m_N": load sharing ratio
--> 28) "sigma_HP": contact strength expressed in [MPa]
--> 29) "Z_N": stress cycle life factor (wear)
--> 30) "Z_W": hardness-ratio factor
--> 31) "wearSF": wear safety factor
Moreover, it can carry out all the necessary calculations.
'''
import numpy as np
from .component import Component
from .force import Force
from math import pi, cos, sin, tan, atan, ceil, sqrt, log
from .makima2dInterpolator import Makima2DInterpolator
class HelicalGear(Component):

    # Overload factor reference matrix
    K0_ref = np.array([[1, 1.25, 1.75], [1.25, 1.5, 2], [1.5, 1.75, 2.25]])
    # Size factor reference graph
    z_ref = np.concatenate((np.arange(12, 23, 1), np.arange(24, 31, 2), np.array([34, 38, 43, 50, 60, 75, 100, 150, 300, 400])))
    Y_ref = np.array([0.245, 0.261, 0.277, 0.29, 0.296, 0.303, 0.309, 0.314, 0.322, 0.328, 0.331, 0.337, 0.346, 0.353, 0.359, 0.371, 0.384, 0.397, 0.409, 0.422, 0.435, 0.447, 0.46, 0.472, 0.48])
    # Bending strength geometry factor reference graphs
    psi_Jp_ref = np.arange(5, 35, 1)
    z_Jp_ref = np.array([20, 30, 60, 150, 500])
    Jp_ref = np.array([[0.465, 0.475, 0.48, 0.487, 0.492, 0.495, 0.497, 0.5, 0.502, 0.505, 0.506, 0.507, 0.508, 0.507, 0.506, 0.505, 0.502, 0.5, 0.497, 0.494, 0.49, 0.486, 0.48, 0.476, 0.471, 0.465, 0.458, 0.452, 0.445, 0.439], 
                       [0.525, 0.53, 0.535, 0.54, 0.542, 0.547, 0.55, 0.552, 0.553, 0.554, 0.555, 0.556, 0.555, 0.554, 0.552, 0.551, 0.549, 0.545, 0.54, 0.537, 0.532, 0.527, 0.52, 0.515, 0.507, 0.5, 0.492, 0.484, 0.476, 0.47], 
                       [0.58, 0.585, 0.595, 0.6, 0.602, 0.605, 0.61, 0.612, 0.614, 0.615, 0.615, 0.615, 0.614, 0.611, 0.608, 0.605, 0.6, 0.595, 0.59, 0.582, 0.577, 0.571, 0.56, 0.555, 0.545, 0.537, 0.527, 0.517, 0.507, 0.495], 
                       [0.62, 0.63, 0.635, 0.64, 0.645, 0.647, 0.652, 0.655, 0.657, 0.657, 0.656, 0.655, 0.654, 0.65, 0.645, 0.64, 0.635, 0.63, 0.622, 0.617, 0.61, 0.6, 0.592, 0.582, 0.575, 0.56, 0.552, 0.541, 0.53, 0.518], 
                       [0.65, 0.655, 0.66, 0.665, 0.67, 0.675, 0.677, 0.68, 0.681, 0.681, 0.681, 0.681, 0.679, 0.676, 0.673, 0.667, 0.66, 0.655, 0.647, 0.64, 0.631, 0.62, 0.613, 0.602, 0.592, 0.58, 0.57, 0.557, 0.545, 0.532]])
    psi_Jpp_ref = np.arange(5, 36, 5)
    z_Jpp_ref = np.array([20, 30, 50, 75, 150, 500])
    Jpp_ref = np.array([[0.927, 0.929, 0.93, 0.932, 0.938, 0.943, 0.953],
                        [0.952, 0.954, 0.957, 0.959, 0.961, 0.965, 0.973],
                        [0.98, 0.981, 0.981, 0.982, 0.983, 0.987, 0.992],
                        np.ones(7),
                        [1.02, 1.02, 1.02, 1.019, 1.017, 1.013, 1.01],
                        [1.038, 1.037, 1.035, 1.032, 1.03, 1.026, 1.02]])
    rel_ref = np.array([.9999, .999, .99, .9, .5])
    YZ_ref = np.array([1.5, 1.25, 1, 0.85, 0.7])
    ZE_ref = np.array([[191, 181, 179, 174, 162, 158], 
                       [181, 174, 172, 168, 158, 154], 
                       [179, 172, 170, 166, 156, 152], 
                       [174, 168, 166, 163, 154, 149], 
                       [162, 158, 156, 154, 145, 141], 
                       [158, 154, 152, 149, 141, 137]])

    # Constructor
    def __init__(self, name, axis, loc, m_n, z, psi, phi_n, Q_v, FW, material):
        # Given parameters
        super().__init__(name=name, material=material, axis=axis, loc=loc, F_tot=None, T_tot=None, omega=None)
        self.m_n = m_n
        self.z = z
        self.psi = psi * pi / 180
        self.phi_n = phi_n * pi / 180
        self.Q_v = Q_v
        self.FW = FW
        # Calculated parameters
        self.p_n = self.m_n * pi
        self.p_t = self.p_n / cos(self.psi)
        self.p_x = self.p_t / tan(self.psi)
        self.m_t = self.m_n / cos(self.psi)
        self.d = self.m_t * self.z
        self.phi_t = atan(tan(self.phi_n) / cos(self.psi))
        self.z_p = ceil(self.z / (cos(self.psi) ** 3))
        self.phi_b = atan(tan(self.psi) * cos(self.phi_n))
        self.h_a = self.m_n
        self.h_f = 1.25 * self.m_n # assumption
        self.h = self.h_a + self.h_f
        self.d_a = self.d + 2 * self.h_a
        self.d_f = self.d - 2 * self.h_f
    
    # Calculate Forces
    def calculateForces(self, mesh, shaft):
        if self.name == mesh.drivingGear.name:
            self.F_t = np.cross(self.T_tot.torque, 2 / self.d * mesh.axis) * 1e3
            mesh.drivingGear.F_t = self.F_t
            magF_t = sqrt(np.sum(self.F_t * self.F_t))
            self.F_r = -magF_t * tan(self.phi_n) / cos(self.psi) * mesh.axis
            mesh.drivingGear.F_r = self.F_r
            self.F_a = np.sign(np.sum(shaft.axis)) * np.sign(self.psi) * np.abs(np.cross(mesh.axis, self.F_t * tan(np.abs(self.psi))))
            mesh.drivingGear.F_a = self.F_a
            Floc = self.d / 2 * np.abs(mesh.axis) + self.loc
            self.F_tot = Force(self.F_t + self.F_r + self.F_a, Floc)
            mesh.drivingGear.F_tot = self.F_tot
        else:
            self.F_t = -mesh.drivingGear.F_t
            self.F_r = -mesh.drivingGear.F_r
            self.F_a = -mesh.drivingGear.F_a
            Floc = -self.d / 2 * np.abs(mesh.axis) + self.loc
            self.F_tot = Force(-mesh.drivingGear.F_tot.force, Floc)
            mesh.drivenGear.F_tot = self.F_tot
    
    # Maximum tooth gear bending stress equation for fatigue
    def calculateSigmaMaxFatigue(self, mesh, powerSource, drivenMachine, dShaft, Ce, teethCond, lShaft, useCond):
        fw_min = min(mesh.drivingGear.FW, mesh.drivenGear.FW)
        # Overload factor K_0
        if powerSource == "Uniform":
            i = 1
        elif powerSource == "Light shock":
            i = 2
        elif powerSource == "Medium shock":
            i = 3
        else:
            raise ValueError("Wrong input for power source")
        if drivenMachine == "Uniform":
            j = 1
        elif drivenMachine == "Moderate shock":
            j = 2
        elif drivenMachine == "Heavy shock":
            j = 3
        else:
            raise ValueError("Wrong input for driven machine")
        self.K_0 = self.__class__.K0_ref[i-1][j-1]
        # Rim-thickness factor
        self.t_R = (self.d_f - dShaft) / 2
        self.m_B = self.t_R / self.h
        if self.m_B < 1.2:
            self.K_B = 1.6 * log(2.242 / self.m_B)
        else:
            self.K_B = 1
        # Dynamic factor
        v = np.sum(np.abs(self.omega)) * self.d * 1e-3
        B = 0.25 * (12 - self.Q_v) ** (2/3)
        A = 50 + 56 * (1 - B)
        self.K_v = ((A + sqrt(200 * v)) / A) ** B
        # Load distribution factor
        self.C_e = Ce
        if teethCond == "uncrowned teeth":
            self.C_mc = 1
        elif teethCond == "crowned teeth":
            self.C_mc = 0.8
        else:
            raise ValueError("Invalid input for teeth condition.")
        if fw_min / 25.4 <= 1:
            self.C_pf = fw_min / 10 / self.d - 0.025
        elif fw_min / 25.4 > 1 and fw_min / 25.4 <= 17:
            self.C_pf = fw_min / 10 / self.d - 0.0375 + 0.0125 * fw_min / 25.4
        else:
            self.C_pf = fw_min / 10 / self.d - 0.1109 + 0.0207 * fw_min / 25.4 - 0.000228 * (fw_min / 25.4) ** 2
        S1 = abs(lShaft / 2 - np.sum(np.abs(self.loc)))
        if S1 / lShaft < 0.175:
            self.C_pm = 1
        else:
            self.C_pm = 1.1
        if useCond == "Open gearing":
            A = 0.247
            B = .0167
            C = -0.765 * 1e-4
        elif useCond == "Commercial, enclosed units":
            A = 0.127
            B = 0.0158
            C = -0.93 * 1e-4
        elif useCond == "Precision, enclosed units":
            A = 0.0675
            B = 0.0128
            C = -0.926 * 1e-4
        elif useCond == "Extraprecision enclosed gear units":
            A = 0.0036
            B = 0.0102
            C = -0.822 * 1e-4
        else:
            raise ValueError("wrong input for use.")
        self.C_ma = A + B * fw_min / 25.4 + C * (fw_min / 25.4) ** 2
        self.K_H = 1 + self.C_mc * (self.C_pf * self.C_pm + self.C_ma * self.C_e)
        # Size factor
        self.Y = np.interp(self.z_p, self.z_ref, self.Y_ref)
        self.K_S = 0.843 * (fw_min * self.m_t * sqrt(self.Y)) ** 0.0535
        # Bending strength geometry factor
        interp_func1 = Makima2DInterpolator(self.psi_Jp_ref, self.z_Jp_ref, self.Jp_ref)
        self.J_p = interp_func1(abs(self.psi * 180 / pi), self.z)
        interp_func2 = Makima2DInterpolator(self.psi_Jpp_ref, self.z_Jpp_ref, self.Jpp_ref)
        self.J_pp = interp_func2(abs(self.psi * 180 / pi), self.z)
        self.Y_J = self.J_p * self.J_pp
        self.sigma_max_fatigue = np.sum(np.abs(self.F_t)) * self.K_0 * self.K_B * self.K_v * self.K_H * self.K_S / fw_min / self.m_t / self.Y_J
    
    # Calulcate bending safety factor
    def calculateBendingSF(self, sigma_FP, b_YN, e_YN, N, temp, rel):
        self.sigma_FP = sigma_FP
        self.Y_N = b_YN * N ** e_YN
        if temp <= 120:
            self.Y_theta = 1
        if rel in self.__class__.rel_ref:
            print("I'm here!")
            ind = np.where(self.__class__.rel_ref == rel)
            self.Y_Z = self.__class__.YZ_ref[ind]
        else:
            self.Y_Z = np.interp(rel, self.__class__.rel_ref, self.__class__.YZ_ref)
        self.bendingSF = self.sigma_FP * self.Y_N / self.sigma_max_fatigue / self.Y_theta / self.Y_Z
    
    # Calculate sigma max pitting
    def calculateSigmaMaxPitting(self, mesh, Z_R):
        # Elastic coefficient
        if mesh.drivingGear.material.name == "Steel":
            i = 1
        elif mesh.drivingGear.material.name == "Malleable iron":
            i = 2
        elif mesh.drivingGear.material.name == "Nodular iron":
            i = 3
        elif mesh.drivingGear.material.name == "Cast iron":
            i = 4
        elif mesh.drivingGear.material.name == "Aluminum bronze":
            i = 5
        elif mesh.drivingGear.material.name == "Tin bronze":
            i = 6
        else:
            raise ValueError("Wrong input for pinion material.")
        if mesh.drivenGear.material.name == "Steel":
            j = 1
        elif mesh.drivenGear.material.name == "Malleable iron":
            j = 2
        elif mesh.drivenGear.material.name == "Nodular iron":
            j = 3
        elif mesh.drivenGear.material.name == "Cast iron":
            j = 4
        elif mesh.drivenGear.material.name == "Aluminum bronze":
            j = 5
        elif mesh.drivenGear.material.name == "Tin bronze":
            j = 6
        else:
            raise ValueError("Wrong input for gear material.")
        self.Z_E = self.__class__.ZE_ref[i-1][j-1]
        # Surface strength geometry
        pG = mesh.drivingGear
        gG = mesh.drivenGear
        f1_ZA = sqrt((pG.d / 2) ** 2 + pG.h_a ** 2)
        f2_Z = (pG.d + gG.d) * sin(self.phi_t) / 2
        Z_A = min([f1_ZA, f2_Z])
        f1_ZB = sqrt((gG.d / 2) ** 2 + gG.h_a ** 2)
        Z_B = min([f1_ZB, f2_Z])
        Z = Z_A + Z_B - f2_Z
        self.m_N = self.p_n * cos(self.phi_n) / 0.95 / Z
        if mesh.type == "External":
            self.Z_I = cos(self.phi_t) * sin(self.phi_t) * mesh.m_G / 2 / self.m_N / (mesh.m_G + 1)
        elif mesh.type == "Internal":
            self.Z_I = cos(self.phi_t) * sin(self.phi_t) * mesh.m_G / 2 / self.m_N / (mesh.m_G - 1)
        else:
            raise ValueError("Gear mesh type invalid.")
        self.sigma_max_pitting = self.Z_E * np.sqrt(np.sum(np.abs(self.F_t)) * self.K_0 * self.K_v * self.K_S * self.K_H * Z_R / self.FW / self.d / self.Z_I)
    # Calculate wear safety factor
    def calculateWearSF(self, sigma_HP, b_ZN, e_ZN, N, mesh):
        self.sigma_HP = sigma_HP
        self.Z_N = b_ZN * N ** e_ZN
        HB_ratio = mesh.drivingGear.material.HB / mesh.drivenGear.material.HB
        if HB_ratio < 1.2:
            A_p = 0
        elif HB_ratio >= 1.2 and HB_ratio <= 1.7:
            A_p = (8.98 * HB_ratio - 8.29) * 1e-3
        else:
            A_p = 0.00698
        self.Z_W = 1 + A_p * (mesh.m_G - 1)
        self.wearSF = self.sigma_HP * self.Z_N * self.Z_W / self.sigma_max_pitting / self.Y_theta / self.Y_Z


'''
This is the "Support" class. It inherits "Component" class.
It defines the properties for SKF bearing supports:
1) "name": a string of characters acting as a label
2) "type": "Pin" / "Roller"
3) "catalogueName": name of bearing as specified in catalogue
4) "catalogueType": "Standard" / "Explorer"
5) "d": bearing internal diamater expressed in [mm]
6) "D": bearing external diameter expressed in [mm]
7) "dm": bearing mean diameter expressed in [mm]
8) "B": bearing width expressed in [mm]
9) "C": dynamic basic load rating expressed in [N]
10) "C0": static basic load rating expressed in [N]
11) "Pu": fatigue load limit expressed in [N]
12) "a": bearing pressure point offset expressed in [mm]
13) "e": calculation factor 1
14) "X": bearing radial load factor
15) "Y": bearing axial load factor
16) "Y0": calculation factor 2
17) "Y1": calculation factor 3
18) "Y2": calculation factor 4
19) "A": bearing calculation factor 1
20) "kr": minimium load factor
21) "shoulder": "1" for +ve offser / "-1" for negative offset
22) "Frm": minimium load expressed in [N]
23) "P": equivalent dynamic bearing load expressed in [N]
24) "P0": equivalent static bearing load expressed in [N]
25) "F_a": axial reaction force expressed in [N]
26) "F_r" radial reaction force expressed in [N]
27) "arr": "Single" / "F2F" / "B2B" / "Tandem" / "Double row"
28) "n": rotational speed expressed in [rpm]
29) "nr": reference speed expressed in [rpm]
30) "s0": safety factor for static bearing load
31) "p": exponent for life equation
32) "a1": reliability factor
33) "eta_c": contamination factor
34) "a_skf": lubrication, contamination, and fatigue limit factor
35) "L_10m": bearing life expressed in [millions of cycles]
36) "L_10mh": bearing life expressed in [hours]
'''
from .component import Component
from .force import Force
import numpy as np
class Support(Component):

    # Reference relibility factor
    rel_ref = np.array([90, 95, 96, 97, 98, 99])
    a1_ref = np.array([1, 0.64, 0.55, 0.47, 0.37, 0.25])

    # Constructor
    def __init__(self, name, type, bearingType, catalogueName, catalogueType, d, D, B, C, C0, Pu, nr, a, e, X, Y, Y0, Y1, Y2, A, kr, shoulder, arr, loc):
        super().__init__(name=name, material=None, axis=None, loc=loc, F_tot=Force(np.array([0, 0, 0]), loc=loc), T_tot=None, omega=None)
        self.type = type
        self.bearingType = bearingType
        self.catalogueName = catalogueName
        self.catalogueType = catalogueType
        self.d = d
        self.D = D
        self.dm = (self.d + self.D) / 2
        self.B = B
        self.C = C
        self.C0 = C0
        self.Pu = Pu
        self.a = a
        self.e = e
        self.X = X
        self.Y = Y
        self.Y0 = Y0
        self.Y1 = Y1
        self.Y2 = Y2
        self.A = A
        self.kr = kr
        self.shoulder = shoulder
        self.nr = nr
        self.n = 0
        self.arr = arr
        if self.bearingType == "ball":
            self.p = 3
        else:
            self.p = 10 / 3
        self.a_skf = 0
    
    # Update reaction force
    def updateReaction(self, shaft):
        for support in shaft.supports:
            if self.name == support.name:
                self.F_tot = support.F_tot
        F_aV = self.F_tot.force * shaft.axis
        self.F_a = np.sqrt(np.sum(F_aV * F_aV))
        F_rV = self.F_tot.force - self.F_a
        self.F_r = np.sqrt(np.sum(F_rV * F_rV))
    
    # Minimum load calculation
    def calculateMinimumLoad(self):
        if self.bearingType == "Tapered":
            if self.catalogueType == "Standard":
                self.Frm = 0.02 * self.C
            elif self.catalogueType == "Explorer":
                self.Frm = 0.017 * self.C
        elif self.bearingType == "Cylindrical":
            self.Frm = self.kr * (6 + 4 * max(self.n) / self.nr) * (self.dm / 100) ** 2 * 1e3
        else:
            raise ValueError("Bearing type not available.")
        if self.F_r != None:
            if self.F_r >= self.Frm:
                print(f"{self.name} satisfies minimium load condition.")
            else:
                print(f"{self.name} does not satisfy minimium load condition.")
    
    # Calculate equivalent dynamic load
    def calculateEquivalentDynamicLoad(self):
        ratioCond = self.F_a / self.F_r <= self.e
        if self.bearingType == "Tapered":
            if self.arr == "Single":
                if ratioCond:
                    self.P = self.F_r
                else:
                    self.P = 0.4 * self.F_r + self.Y * self.F_a
            elif self.arr == "F2F":
                if ratioCond:
                    self.P = self.F_r + self.Y1 * self.F_a
                else:
                    self.P = 0.67 * self.F_r + self.Y2 * self.F_a
            elif self.arr == "B2B":
                if ratioCond:
                    self.P = self.F_r + self.Y1 * self.F_a
                else:
                    self.P = 0.67 * self.F_r + self.Y2 * self.F_a
            elif self.arr == "Tandem":
                if ratioCond:
                    self.P = self.F_r
                else:
                    self.P = 0.4 * self.F_r + self.Y * self.F_a
            elif self.arr == "Double row":
                if ratioCond:
                    self.P = self.F_r + self.Y1 * self.F_a
                else:
                    self.P = 0.67 * self.F_r + self.Y2 * self.F_a
            else:
                raise ValueError("Bearing arrangement not available.")
        elif self.bearingType == "Cylindrical":
            if ratioCond:
                self.P = self.F_r
            else:
                self.P = 0.92 * self.F_r + self.Y * self.F_a
        else:
            raise ValueError("Bearing type not available.")
    
    # Calculate equivalent static load
    def calculateEquivalentStaticLoad(self):
        if self.bearingType == "Tapered":
            if self.arr == "Single":
                self.P0 = 0.5 * self.F_r + self.Y0 * self.F_a
                if self.P0 < self.F_r:
                    self.P0 = self.F_r
            elif self.arr == "F2F":
                self.P0 = self.F_r + self.Y0 * self.F_a
                if self.P0 < self.F_r:
                    self.P0 = self.F_r
            elif self.arr == "B2B":
                self.P0 = self.F_r + self.Y0 * self.F_a
                if self.P0 < self.F_r:
                    self.P0 = self.F_r
            elif self.arr == "Tandem":
                self.P0 = 0.5 * self.F_r + self.Y0 * self.F_a
            elif self.arr == "Double row":
                self.P0 = self.F_r + self.Y0 * self.F_a
                if self.P0 < self.F_r:
                    self.P0 = self.F_r
            else:
                raise ValueError("Bearing arrangement not avaialabe.")
        if self.bearingType == "Cylindrival":
            self.P0 = self.F_r
        self.s0 = self.C0 / self.P0
    
    # Calculate a1
    def calculateA1(self, rel):
        self.a1 = np.interp(rel, self.__class__.rel_ref, self.__class__.a1_ref)
    
    # Calculate contamination factor
    def calculateEtaC(self, condition):
        if condition == "Extreme cleanliness":
            self.eta_c = 1
        elif condition == "High cleanliness":
            if self.dm < 100:
                self.eta_c = 0.7
            else:
                self.eta_c = 0.85
        elif condition == "Normal cleanliness":
            if self.dm < 100:
                self.eta_c = 0.55
            else:
                self.eta_c = 0.7
        elif condition == "Slight contamination":
            if self.dm < 100:
                self.eta_c = 0.4
            else:
                self.eta_c = 0.5
        elif condition == "Typical contamination":
            if self.dm < 100:
                self.eta_c = 0.2
            else:
                self.eta_c = 0.3
        elif condition == "Severe contamination":
            self.eta_c = 0.05
        elif condition == "Very severe contamination":
            self.eta_c = 0
        else:
            raise ValueError("Invalid contamination condition.")

    # Calculate bearing life
    def calculateBearingLife(self):
        self.L_10m = self.a1 * self.a_skf * (self.C / self.P) ** self.p
        self.L_10mh = 1e6 / 60 / np.abs(self.n[np.where(self.n != 0)]) * self.L_10m


'''
This is the "Oil" class.
It is used to define the oil properties of the oil being used.
The properties are:
1) "name": a string of characters acting as a label
2) "temperature": the oil operating temperature expressed in [degC]
3) "v1": Rated viscosity of the oil expressed in [mm^2/s]
4) "v": viscosity of the oil expressed in [mm^2/s]
5) "k": viscosity ratio of the oil
'''
class Oil:

    # Constructor
    def __init__(self, name, temp, v1, v):
        self.name = name
        self.temperature = temp
        self.v1 = v1
        self.v = v
        self.l = self.v / self.v1


'''
This is the "Force" class.
It defines a force vector based on two simple properties:
1) "force": a 3-element force vector representing the force expressed in [N]
2) "loc": a 3-element vector representing the point of application of the force expressed in [mm]
'''
class Force:

    # Constructor
    def __init__(self, force, loc):
        self.force = force
        self.loc = loc
    
    # Overload Addition
    def __add__(obj1, obj2):
        return obj1.force + obj2.force
    
    # Overload Subtraction
    def __sub__(obj1, obj2):
        return obj1.force - obj2.force
    
    # Overload Negative
    def __neg__(obj):
        return -obj.force
    
    # Overload Equal
    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return all(self.force == other.force) and all(self.loc == other.loc)
        return False
    
    # Overload Call
    def __call__(self):
        print(f"Force: {self.force}\nLoc: {self.loc}")


'''
This is the "ShaftSection" class.
It's used to define certain sections of a "Shaft" component along its profile.
It has the following properties:
I) Given Properties:
--> 1) "name": a string of characters acting as a label
--> 2) "loc": a 3-D element vector representing the location of the section along the shaft
--> 3) "d": a scalar representing the diameter of the shaft at the specified cross-section expressed in [mm]
--> 4) "Ra": a scalar representing the surface roughness at the specified cross-section expressed in [micrometers]
--> 5) "material": a Material object representing the material of the shaft
II) Calculated Properties:
--> 1) "sigma_m_Mb": mean bending stress [MPa]
--> 2) "sigma_a_Mb": alternating bending stress [MPa]
--> 3) "sigma_m_N": mean normal stress [MPa]
--> 4) "sigma_a_N": alternating normal stress [MPa]
--> 5) "tau_m_Mt": mean torsional stress [MPa]
--> 6) "tau_a_Mt": alternating torsional stress [MPa]
--> 7) "Kt_B": bending stress concentration rasiser
--> 8) "Kt_N": normal stress concentration raiser
--> 9) "Kt_T": torsion stress concentration raiser
--> 10) "Kf_B": bending stress concentration factor
--> 11) "Kf_N": normal stress concentration factor
--> 12) "Kf_T": torsion stress concentration factor
'''
from .notchSensitivity import NotchSensitivity
from .fatigueLimitCorrectorFactors import FatigueLimitCorrectorFactors
from .geometricStressRaiser import GeometricStressRaiser
from math import sqrt
import numpy as np
import matplotlib.pyplot as plt
class ShaftSection:

    # Constructor
    def __init__(self, name, loc, d, Ra, material):
        self.name = name
        self.loc = loc
        self.d = d
        self.Ra = Ra
        self.material = material
        # initialize mean and alternating stresses to 0
        self.sigma_m_Mb = 0
        self.sigma_a_Mb = 0
        self.sigma_m_N = 0
        self.sigma_a_N = 0
        self.tau_m_Mt = 0
        self.tau_a_Mt = 0
        # initialize stress concentration raisers
        self.Kt_B = 0
        self.Kt_N = 0
        self.Kt_T = 0
        # initialize stress concentration factor
        self.Kf_N = 1
        self.Kf_B = 1
        self.Kf_T = 1
        # initialize static safety factor
        self.staticSF = 0
    
    # Append fatigue stress intensification factor
    def AppendKf(self, Kf, loadType):
        for kf, lt in zip(Kf, loadType):
            if lt == "Normal":
                self.Kf_N = self.Kf_N * kf
            elif lt == "Bending":
                self.Kf_B = self.Kf_B * kf
            elif lt == "Torsion":
                self.Kf_T = self.Kf_T * kf
            else:
                raise ValueError(f"'{lt}' is Invalid")
            
    # Calculate fatigue stress intensification factor
    def CalculateFatigueIntensificationFactor(self):
        self.__class__.AppendKf(self, [1 + self.q.qReq * (self.Kt_B - 1)], ["Bending"])
        self.__class__.AppendKf(self, [1 + self.q.qReq * (self.Kt_N - 1)], ["Normal"])
        self.__class__.AppendKf(self, [1 + self.q.qReq * (self.Kt_T - 1)], ["Torsion"])

    # Notch Sensitivity
    def AddNotchSensitivity(self, notchRadius, sigma_u):
        self.q = NotchSensitivity(notchRadius=notchRadius, sigma_u=sigma_u)
        if self.Kt_B == None and self.Kt_N == None and self.Kt_T == None:
            self.__class__.CalculateFatigueIntensificationFactor(self)
    
    #Geomtric Stress Raiser
    def AddGeometricStressRaiser(self, r2d, D2d):
        temp = GeometricStressRaiser(r2d, D2d)
        self.Kt_B = temp.Kt_Breq
        self.Kt_N = temp.Kt_Nreq
        self.Kt_T = temp.Kt_Treq
        self.CalculateFatigueIntensificationFactor()
    
    # Add fatigue limit corrector factors
    def AddFLCF(self):
        self.FLCF = FatigueLimitCorrectorFactors(self)
        self.material.ComponentFatigueLimit(self)
    
    # Calculate Equivalent Stress
    def CalculateEquivalentStress(self):
        self.sigma_a_eq = sqrt((self.Kf_B * self.sigma_a_Mb + self.Kf_N * self.sigma_a_N/0.85) ** 2 + 3 * (self.Kf_T * self.tau_a_Mt) ** 2)
        self.sigma_m_eq = sqrt((self.Kf_B * self.sigma_m_Mb + self.Kf_N * self.sigma_m_N/0.85) ** 2 + 3 * (self.Kf_T * self.tau_m_Mt) ** 2)

    # Plot Haigh Diagram
    def PlotHaighDiagram(self):
        # Calculate coordinates of specific points
        coeff1 = np.polyfit(np.array([0, self.material.sigma_y]), np.array([self.material.sigma_y, 0]), deg=1)
        coeff2 = np.polyfit(np.array([0, self.material.sigma_u]), np.array([self.material.sigma_Dm1C, 0]), deg=1)
        inter_x = (coeff2[1] - coeff1[1]) / (coeff1[0] - coeff2[0])
        inter_y = np.polyval(coeff1, inter_x)
        coeff3 = np.polyfit(np.array([0, self.sigma_m_eq]), np.array([0, self.sigma_a_eq]), deg=1)
        P_x = (coeff2[1] - coeff3[1]) / (coeff3[0] - coeff2[0])
        if P_x > inter_x:
            P_x = (coeff1[1] - coeff3[1]) / (coeff3[0] - coeff1[0])
        P_y = np.polyval(coeff3, P_x)
        # Plot
        plt.figure
        plt.plot(np.array([0, inter_x, self.material.sigma_y]), np.array([self.material.sigma_Dm1C, inter_y, 0]), 'k', linewidth=1.5)
        plt.plot(self.sigma_m_eq, self.sigma_a_eq, 'm*', linewidth=1.5)
        plt.plot(np.array([0, inter_x, self.material.sigma_u]), np.array([self.material.sigma_y, inter_y, 0]), 'k--', linewidth=1.5)
        plt.plot(0, self.material.sigma_Dm1C, "bo", linewidth=1.5)
        plt.plot(0, self.material.sigma_y, "bo", linewidth=1.5)
        plt.plot(self.material.sigma_y, 0, "bo", linewidth=1.5)
        plt.plot(self.material.sigma_u, 0, "bo", linewidth=1.5)
        plt.plot(0, P_y, "bo", linewidth=1.5)
        plt.plot(np.array([0, P_x]), np.array([0, P_y]), "g--", linewidth=1.5)
        plt.plot(np.array([0, P_x]), np.array([P_y, P_y]), "g--", linewidth=1.5)
        plt.text(self.sigma_m_eq, self.sigma_a_eq - 35, "P" + self.name[1])
        plt.text(10, self.material.sigma_Dm1C + 15, r"$\sigma_{D-1}^{C}$")
        plt.text(self.material.sigma_y, 20, r"$\sigma_{y}$")
        plt.text(10, self.material.sigma_y + 15, r"$\sigma_{y}$")
        plt.text(self.material.sigma_u, 20, r"$\sigma_{u}$")
        plt.text(10, P_y - 35, r"$\sigma_{D, lim}$")
        plt.xlabel(r"Mean Stress - $\sigma_{m, eq}$ [MPa]")
        plt.ylabel(r"Alternating Stress - $\sigma_{a, eq}$ [MPa]")
        plt.title("Haigh Diagram @ " + self.name)
        plt.grid()
        plt.show()
    
    # Calculate Fatigue Safety Factor
    def CalculateFatigueSF(self):
        # Calculate coordinates of specific points
        coeff1 = np.polyfit(np.array([0, self.material.sigma_y]), np.array([self.material.sigma_y, 0]), deg=1)
        coeff2 = np.polyfit(np.array([0, self.material.sigma_u]), np.array([self.material.sigma_Dm1C, 0]), deg=1)
        inter_x = (coeff2[1] - coeff1[1]) / (coeff1[0] - coeff2[0])
        inter_y = np.polyval(coeff1, inter_x)
        coeff3 = np.polyfit(np.array([0, self.sigma_m_eq]), np.array([0, self.sigma_a_eq]), deg=1)
        P_x = (coeff2[1] - coeff3[1]) / (coeff3[0] - coeff2[0])
        if P_x > inter_x:
            P_x = (coeff1[1] - coeff3[1]) / (coeff3[0] - coeff1[0])
        P_y = np.polyval(coeff3, P_x)
        self.fatigueSF = P_y / self.sigma_a_eq


__version__ = "0.1.2"

from .component import Component
from .fatigueLimitCorrectorFactors import FatigueLimitCorrectorFactors
from .force import Force
from .gearMesh import GearMesh
from .geometricStressRaiser import GeometricStressRaiser
from .helicalGear import HelicalGear
from .inputMotor import InputMotor
from .makima2dInterpolator import Makima2DInterpolator
from .material import Material
from .notchSensitivity import NotchSensitivity
from .oil import Oil
from .shaft import Shaft
from .shaftProfile import ShaftProfile
from .shaftSection import ShaftSection
from .support import Support
from .torque import Torque


'''
This is the "Component" class.
It's a parent class that defines general properties that are common among different components.
Other classes will inherit these properties instead of having to define every single time.
Properties are:
1) "name": a string of characters acting as a label
2) "material": a "Material" object of the material class defining the material properties of the component
3) "axis": a 3-element vector representing the axis along which the component is rotating with respect to a defined reference frame
4) "loc": a 3-element vector representing the location of the component with respect to a defined reference frame
5) "F_tot": the total force acting on the component expressed in [N]
6) "T_tot": the total torque acting on the component expressed in [Nm]
'''
class Component:

    # Constructor
    def __init__(self, name, material, axis, loc, F_tot, T_tot, omega):
        self.name = name
        self.material = material
        self.axis = axis
        self.loc = loc
        self.F_tot = F_tot
        self.T_tot = T_tot
        self.omega = omega


'''
This is the "Material" class.
It is used to define the material properties for the components.
The properties are:
1) "name": a string of characters acting as a label
2) "sigma_u": Ultimate tensile strength expressed in [MPa]
3) "sigma_y": Yielding strength expressed in [MPa]
4) "sigma_Dm1": Material specific fatigue limit expressed in [MPa]
5) "HB": Hardness of the material
6) "sigma_Dm1C": Component specific fatigue limit expressed in [MPa]
'''
class Material:

    # Constructor
    def __init__(self, name, sigma_u, sigma_y, sigma_Dm1, HB):
        self.name = name
        self.sigma_u = sigma_u
        self.sigma_y = sigma_y
        self.sigma_Dm1 = sigma_Dm1
        self.HB = HB
    
    # Component Fatigue Limit
    def ComponentFatigueLimit(self, section):
        self.sigma_Dm1C = self.sigma_Dm1*section.FLCF.C
        return self


'''
This is the gearMesh class.
It defines a gear mesh by relying on 2 gear components and the type of mesh.
The properties used are:
I) Given properties
--> 1) "drivingGear": a gear component to which the input is connected to
--> 2) "drivenGear": a gear component to which the output in connectec to
--> 3) "axis": a 3-element vector representing the axis of rotation
--> 4) "type": a string of characters indicating the gear mesh type: Internal / External
II) Calculated properties
--> 1) "ratio": the gear ratio calculated based on the number of teeth
--> 2) "m_G": the reciprocal of the gear ratio
After creating an instance of the class, it's possible to get the rotational velocity as well as the torque corresponding to the driven gear via:
1) GetOmegaMesh()
2) GetMeshTorque()
methods, respectively.
'''
from .torque import Torque
class GearMesh:

    # Constructor
    def __init__(self, drivingGear, drivenGear, axis, type):
        if drivingGear.m_n != drivenGear.m_n:
            raise Exception("Incompatible Gear Mesh!")
        # Given properties
        self.drivingGear = drivingGear
        self.drivenGear = drivenGear
        self.axis = axis
        self.type = type
        # Calculated properties
        self.ratio = self.drivingGear.z / self.drivenGear.z
        self.m_G = 1 / self.ratio
        sgn = -1 # aassuming self.type = "External"
        if self.type == "Internal":
            sgn = 1
        self.drivenGear.omega = sgn * self.ratio * self.drivingGear.omega
        self.drivenGear.T_tot = Torque(-sgn * self.drivingGear.T_tot.torque / self.ratio, self.drivenGear.loc)
    
    # Get Driven Gear Omega
    def GetOmegaMesh(self):
        return self.drivenGear.omega
    
    # Get Driven Gear Torque
    def GetMeshTorque(self):
        return self.drivenGear.T_tot.torque


from setuptools import find_packages, setup

with open("README.md", 'r') as f:
    long_description = f.read()

setup(
    name="pygritbx",
    version="0.1.2",
    description='Python-based Gearbox Reliability and Integrity Tool',
    #package_dir={"": "pygrit"},
    packages=find_packages(),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Rmhuneineh/pygrit",
    author="Ragheed Huneineh",
    author_email="ragheedmhuneineh@outlook.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.11",
        "Operating System :: OS Independent",
        "Intended Audience :: Education",
        "Natural Language :: English",
        "Topic :: Scientific/Engineering",
    ],
    install_requires=[
        "numpy >= 2.2.4",
        "scipy >= 1.15.2"
    ],
    extras_require={
        "dev": ["twine >= 6.1.0"]
    },
    python_requires=">= 3.11.5",
)


'''
This is the FatigueLimitCorrectorFactors class.
It contains the data and the necessary calculations via interpolation to get the final factor.
The properties are:
1) "Cs": size factor
2) "Cl": load factor
3) "Cf": surface finish factor
'''
import numpy as np
from scipy.interpolate import RegularGridInterpolator
class FatigueLimitCorrectorFactors:
    
    # CONSTANTS
    # Graph reference diameter - [mm]
    d_ref = np.arange(10, 201, 10)
    # Reference size factor - [-]
    Cs_ref = np.array([1, 0.93, 0.875, 0.845, 0.82, 0.8, 0.785, 0.773, 0.76, 0.753, 
                       0.745, 0.74, 0.73, 0.725, 0.72, 0.715, 0.71, 0.705, 0.7, 0.695])
    # Reference load factor - [-]
    Cl_ref = 0.7
    # Reference ultimate tensile strength - [MPa]
    sigma_u_ref = np.arange(200, 1501, 100)
    # Reference roughness - [micrometers]
    Ra_ref = np.array([0.8, 1.6, 3.2, 6.3, 10, 40, 160])
    # Reference surface finish factor - [-]
    Cf_ref = np.array([np.ones(14), 
                       [1, 0.98, 0.97, 0.96, 0.96, 0.96] + [0.95]*8,
                       [1, 0.97, 0.96, 0.94, 0.93, 0.93, 0.92] + [0.91]*7,
                       [0.98, 0.95, 0.93, 0.92, 0.91, 0.9, 0.89, 0.88, 0.87] + [0.86]*5,
                       [0.97, 0.95, 0.92, 0.88, 0.87, 0.84, 0.82, 0.81, 0.78, 0.77, 0.76] + [0.75]*3,
                       [0.96, 0.92, 0.88, 0.84, 0.8, 0.76, 0.73, 0.7, 0.68, 0.66, 0.65, 0.63, 0.63, 0.62],
                       [0.94, 0.89, 0.85, 0.8, 0.76, 0.72, 0.67, 0.65, 0.61, 0.58, 0.55, 0.52, 0.5, 0.48]])
    # Interpolation Function
    interp_func = RegularGridInterpolator((Ra_ref, sigma_u_ref), Cf_ref)
    # Constructor
    def __init__(self, section):
        # Cf
        self.Cf_req = self.__class__.interp_func([[section.Ra, section.material.sigma_u]])[0]
        # Cl
        if section.sigma_m_N != 0 or section.sigma_a_N != 0:
            self.Cl_req = self.__class__.Cl_ref
        else:
            self.Cl_req = 0
        # Cs
        if section.sigma_m_Mb != 0 or section.sigma_a_Mb != 0 or section.sigma_m_Mt != 0 or section.sigma_a_Mt != 0:
            self.Cs_req = np.interp(section.d, self.__class__.d_ref, self.__class__.Cs_ref)
        else:
            self.Cs_req = 1
        # C
        self.C = self.Cf_req * self.Cl_req * self.Cs_req